/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2005 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "OgreOSXWindow.h"
#include "OgreRoot.h"
#include "OgreRenderSystem.h"
#include "OgreImageCodec.h"
#include "OgreException.h"
#include "OgreLogManager.h"
#include "OgreStringConverter.h"

#include "OgreGLRenderSystem.h"
#include "OgreGLOSXContext.h"

#include <OpenGL/gl.h>
#define GL_EXT_texture_env_combine 1
#include <OpenGL/glext.h>
#include <OpenGL/glu.h>


@implementation OgreView

- (id)initWithFrame:(NSRect)f
{
	[super initWithFrame:f];
	
	window = 0;
	keyEvents = [[NSMutableArray alloc] init];
	mouseEvents = [[NSMutableArray alloc] init];
	
	return self;
}

- (id)initWithGLOSXWindow:(Ogre::OSXWindow*)w
{
	[super init];
	
	keyEvents = [[NSMutableArray alloc] init];
	mouseEvents = [[NSMutableArray alloc] init];

	window = w;
	return self;
}

- (void)setOgreWindow:(Ogre::OSXWindow*)w
{
	window = w;
}

- (Ogre::OSXWindow*)ogreWindow
{
	return window;
}

- (void)setFrameSize:(NSSize)s
{
	[super setFrameSize:s];
	window->windowResized();
}

- (NSMutableArray*)keyEvents {return keyEvents;}
- (NSMutableArray*)mouseEvents {return mouseEvents;}

- (void)keyDown:(NSEvent *)theEvent
{
	[keyEvents addObject:theEvent];
}

- (void)keyUp:(NSEvent *)theEvent
{
	[self keyDown:theEvent];
}

- (void)flagsChanged:(NSEvent *)theEvent
{
	[self keyDown:theEvent];
}

- (void)mouseDown:(NSEvent *)theEvent
{
	[mouseEvents addObject:theEvent];
}

- (void)rightMouseDown:(NSEvent *)theEvent
{
	[self mouseDown:theEvent];
}

- (void)otherMouseDown:(NSEvent *)theEvent
{
	[self mouseDown:theEvent];
}

- (void)mouseUp:(NSEvent *)theEvent
{
	[self mouseDown:theEvent];
}

- (void)rightMouseUp:(NSEvent *)theEvent
{
	[self mouseDown:theEvent];
}

- (void)otherMouseUp:(NSEvent *)theEvent
{
	[self mouseDown:theEvent];
}

- (void)scrollWheel:(NSEvent *)theEvent
{
	[self mouseDown:theEvent];
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- (void)drawRect:(NSRect)r
{
	if(window)
		window->update();
}

@end


namespace Ogre {

    OSXWindow::OSXWindow()
    {
		mActive = mClosed = mHasResized = false;
		mContext = 0;
		mWindow = 0;
    }

    OSXWindow::~OSXWindow()
    {
		[glContext clearDrawable];
		CGReleaseAllDisplays();
        //[mWindow close];
		//[mWindow release];
    }
	
	void OSXWindow::create(const String& name, unsigned int width, unsigned int height,
	            bool fullScreen, const NameValuePairList *miscParams)
    {
		//OgreWindowDelegate *delegate = [[OgreWindowDelegate alloc] initWithGLOSXWindow:this];
		//[window setDelegate:delegate];
		/*
***Key: "title" Description: The title of the window that will appear in the title bar Values: string Default: RenderTarget name

Key: "colourDepth" Description: Colour depth of the resulting rendering window; only applies if fullScreen is set. Values: 16 or 32 Default: desktop depth Notes: [W32 specific]

***Key: "left" Description: screen x coordinate from left Values: positive integers Default: 'center window on screen' Notes: Ignored in case of full screen

***Key: "top" Description: screen y coordinate from top Values: positive integers Default: 'center window on screen' Notes: Ignored in case of full screen

Key: "depthBuffer" [DX9 specific] Description: Use depth buffer Values: false or true Default: true

***Key: "externalWindowHandle" [API specific] Description: External window handle, for embedding the OGRE context Values: positive integer for W32 (HWND handle) poslong:posint:poslong (display*:screen:windowHandle) or poslong:posint:poslong:poslong (display*:screen:windowHandle:XVisualInfo*) for GLX Default: 0 (None)

***Key: "FSAA" Description: Full screen antialiasing factor Values: 0,2,4,6,... Default: 0

Key: "displayFrequency" Description: Display frequency rate, for fullscreen mode Values: 60...? Default: Desktop vsync rate

Key: "vsync" Description: Synchronize buffer swaps to vsync Values: true, false Default: 0
*/
		BOOL hasDepthBuffer;
		int fsaa_samples;
		NSString *windowTitle;
		int winx, winy;
		
		if(miscParams)
		{
			NameValuePairList::const_iterator opt = 0;
			
			opt = miscParams->find("title");
			if(opt != miscParams->end())
				windowTitle = [NSString stringWithCString:opt->second.c_str()];
				
			opt = miscParams->find("left");
			if(opt != miscParams->end())
				winx = StringConverter::parseUnsignedInt(opt->second);
				
			opt = miscParams->find("top");
			if(opt != miscParams->end())
				winy = NSHeight([[NSScreen mainScreen] frame]) - StringConverter::parseUnsignedInt(opt->second) - height;
				
				
			opt = miscParams->find("depthBuffer");
			if(opt != miscParams->end())
				hasDepthBuffer = StringConverter::parseBool(opt->second);
				
			opt = miscParams->find("FSAA");
			if(opt != miscParams->end())
				fsaa_samples = StringConverter::parseUnsignedInt(opt->second);
		}
		else
		{
			fsaa_samples = 0;
			hasDepthBuffer = YES;
			winx = winy = 0;
			windowTitle = [NSString stringWithCString:name.c_str()];
		}
		
		

		NSOpenGLPixelFormat* openglFormat;
		{
			NSOpenGLPixelFormatAttribute attribs[20];
			int i=0;
			
			if(fullScreen)
			{
				// Specify that we want a full-screen OpenGL context.
				attribs[i++] = NSOpenGLPFAFullScreen;
				
				// We may be on a multi-display system (and each screen may be driven by a different renderer), so we need to specify which screen we want to take over.  For this demo, we'll specify the main screen.
				attribs[i++] = NSOpenGLPFAScreenMask;
				attribs[i++] = (NSOpenGLPixelFormatAttribute)CGDisplayIDToOpenGLDisplayMask(kCGDirectMainDisplay);
			}
			else
			{
				// Specifying "NoRecovery" gives us a context that cannot fall back to the software renderer.  This makes the View-based context a compatible with the fullscreen context, enabling us to use the "shareContext" feature to share textures, display lists, and other OpenGL objects between the two.
				attribs[i++] = NSOpenGLPFANoRecovery;
			}
			
			attribs[i++] = NSOpenGLPFAAccelerated;
			attribs[i++] = NSOpenGLPFADoubleBuffer;
			
			attribs[i++] = NSOpenGLPFAAlphaSize;
			attribs[i++] = (NSOpenGLPixelFormatAttribute) 0;
			
			attribs[i++] = NSOpenGLPFAStencilSize;
			attribs[i++] = (NSOpenGLPixelFormatAttribute) 1;
			
			attribs[i++] = NSOpenGLPFAAccumSize;
			attribs[i++] = (NSOpenGLPixelFormatAttribute) 0;
			
			attribs[i++] = NSOpenGLPFADepthSize;
			attribs[i++] = (NSOpenGLPixelFormatAttribute) (hasDepthBuffer? 16 : 0);
			
			if(fsaa_samples > 0)
			{
				attribs[i++] = NSOpenGLPFASampleBuffers;
				attribs[i++] = (NSOpenGLPixelFormatAttribute) 1;
				
				attribs[i++] = NSOpenGLPFASamples;
				attribs[i++] = (NSOpenGLPixelFormatAttribute) fsaa_samples;
			}
			
			attribs[i++] = (NSOpenGLPixelFormatAttribute) 0;

			openglFormat = [[[NSOpenGLPixelFormat alloc] initWithAttributes: attribs] autorelease];
		}

		GLRenderSystem *rs = static_cast<GLRenderSystem*>(Root::getSingleton().getRenderSystem());
		GLOSXContext *mainContext = (GLOSXContext*)rs->_getMainContext();
		NSOpenGLContext *shareContext = mainContext == 0? nil : mainContext->getContext();

		glContext = [[NSOpenGLContext alloc] initWithFormat: openglFormat shareContext:shareContext];
		
		NameValuePairList::const_iterator opt = 0;
		if(miscParams)
			opt = miscParams->find("externalWindowHandle");
		if(!miscParams || opt == miscParams->end())
		{
			//Not sure why this should be but it is required for the window to work at fullscreen.
			int styleMask = fullScreen? NSBorderlessWindowMask : NSResizableWindowMask;
		
			mWindow = [[NSWindow alloc] initWithContentRect:NSMakeRect(winx, winy, width, height) styleMask:styleMask backing:NSBackingStoreBuffered defer:NO];
			[mWindow setTitle:windowTitle];
			
			if(winx == 0 && winy == 0) [mWindow center];
			
			mView = [[OgreView alloc] initWithGLOSXWindow:this];
			[mWindow setContentView:mView];
		}
		else
		{
			mView = (OgreView*)StringConverter::parseUnsignedLong(opt->second);
			[mView setOgreWindow:this];
			
			NSRect b = [mView bounds];
			width = b.size.width;
			height = b.size.height;
		}
	
		if(fullScreen)
		{
			CGDisplayFadeReservationToken reservationToken;
			CGAcquireDisplayFadeReservation(kCGMaxDisplayReservationInterval,
											&reservationToken);
			CGDisplayFade(reservationToken,
						  0.5,
						  kCGDisplayBlendNormal,
						  kCGDisplayBlendSolidColor,
						  0.0, 0.0, 0.0,
						  true);
		
			CGDisplayCapture( kCGDirectMainDisplay );
			CGDisplaySwitchToMode( kCGDirectMainDisplay,
				CGDisplayBestModeForParameters( kCGDirectMainDisplay,
					32, width, height, 0 ) ) ;
			
			[glContext setFullScreen];
			[NSCursor hide];
			
			CGDisplayFade(reservationToken,
                  2.0,
                  kCGDisplayBlendSolidColor,
                  kCGDisplayBlendNormal,
                  0.0, 0.0, 0.0,
                  false);
			CGReleaseDisplayFadeReservation(reservationToken);
		}
		else
		{
			[glContext setView:mView];
		}

        mName = name;
        mWidth = width;
        mHeight = height;
		
		// Create register the context with the rendersystem and associate it with this window
		mContext = new GLOSXContext(glContext);
		/*rs->_registerContext(this, newContext);
		
		if (rs->_getMainContext() == 0)
			[glContext makeCurrentContext];
		*/
		//show and make active
		if(mWindow)
			[mWindow performSelectorOnMainThread:@selector(makeKeyAndOrderFront:) withObject:NULL waitUntilDone:NO];
		mActive = true;
    }

    void OSXWindow::destroy(void)
    {
		// Unregister and destroy OGRE GLContext
		delete mContext;
		
		if(mWindow) [mWindow close];
		
        mActive = false;

        Root::getSingleton().getRenderSystem()->detachRenderTarget( this->getName() );
    }

	/*OgreView* OSXWindow::ogreView() const
	{
		return [window contentView];
	}*/

    bool OSXWindow::isActive() const
    {
        return mActive;
    }

    bool OSXWindow::isClosed() const
    {
        return false;
    }

    void OSXWindow::reposition(int left, int top)
    {
		if(!mWindow) return;
		
		NSRect frame = [mWindow frame];
		frame.origin.x = left;
		frame.origin.y = top-frame.size.height;
		[mWindow setFrame:frame display:YES];
    }

    void OSXWindow::resize(unsigned int width, unsigned int height)
    {
		if(!mWindow) return;
		
		NSRect frame = [mWindow frame];
		frame.size.width = width;
		frame.size.height = height;
		[mWindow setFrame:frame display:YES];
	/*SHOULD be called from delegate
        for (ViewportList::iterator it = mViewportList.begin(); it != mViewportList.end(); ++it)
        {
            (*it).second->_updateDimensions();
        }*/
    }

    void OSXWindow::windowHasResized()
    {
		mHasResized = true;
		[glContext update];
    }

    void OSXWindow::windowResized()
    {
		[glContext update];
		NSRect frame = [mView frame];
		mWidth = (unsigned int)frame.size.width;
		mHeight = (unsigned int)frame.size.height;
        mLeft = (int)frame.origin.x;
        mTop = (int)frame.origin.y+(unsigned int)frame.size.height;
		
        for (ViewportList::iterator it = mViewportList.begin(); it != mViewportList.end(); ++it)
        {
            (*it).second->_updateDimensions();
        }
    }

    void OSXWindow::swapBuffers(bool waitForVSync)
    {
		[glContext flushBuffer];
		if(mHasResized)
		{
			windowResized();
			mHasResized = false;
		}
		/*if(![[[NSRunLoop currentRunLoop] currentMode] isEqualTo:NSEventTrackingRunLoopMode])
		{
			NSEvent *e = [NSApp nextEventMatchingMask:NSAnyEventMask untilDate:[NSDate distantPast] inMode:NSDefaultRunLoopMode dequeue:YES];
			if(e)
				[NSApp sendEvent:e];
		}*/
    }
	
	//-------------------------------------------------------------------------------------------------//
	void OSXWindow::getCustomAttribute( const String& name, void* pData )
	{
		if( name == "GLCONTEXT" ) 
		{
			*static_cast<GLOSXContext**>(pData) = mContext;
			return;
		} 
	}

	void OSXWindow::writeContentsToFile(const String& filename)
	{
		ImageCodec::ImageData* imgData = new ImageCodec::ImageData;
		imgData->width = mWidth;
		imgData->height = mHeight;
		imgData->format = PF_BYTE_RGB;

		// Allocate buffer 
		uchar* pBuffer = new uchar[mWidth * mHeight * 3];

		// Read pixels
		// I love GL: it does all the locking & colour conversion for us
		glReadPixels(0,0, mWidth-1, mHeight-1, GL_RGB, GL_UNSIGNED_BYTE, pBuffer);

		// Wrap buffer in a memory stream
        DataStreamPtr stream(new MemoryDataStream(pBuffer, mWidth * mHeight * 3, false));

		// Need to flip the read data over in Y though
		Image img;
		img.loadRawData(stream, mWidth, mHeight, imgData->format );
		img.flipAroundX();

        MemoryDataStreamPtr streamFlipped(new MemoryDataStream(img.getData(), stream->size(), false));

		// Get codec 
		size_t pos = filename.find_last_of(".");
		String extension;
		if( pos == String::npos )
			OGRE_EXCEPT(
			Exception::ERR_INVALIDPARAMS, 
			"Unable to determine image type for '" + filename + "' - invalid extension.",
			"OSXWindow::writeContentsToFile" );

		while( pos != filename.length() - 1 )
			extension += filename[++pos];

		// Get the codec
		Codec * pCodec = Codec::getCodec(extension);

		// Write out
		Codec::CodecDataPtr codecDataPtr(imgData);
		pCodec->codeToFile(streamFlipped, filename, codecDataPtr);

		delete [] pBuffer;
	}
}
